use std::iter::once;

use crate::common::SectionInfo;
use crate::common::{Relocatable, Section};
use crate::error::Error;

pub enum SegmentType {
    Text,
    Data,
    Bss,
}

#[derive(Default)]
pub struct SegmentSections {
    sections_info: Vec<SectionInfo>,
    data_size: u64,
}

impl SegmentSections {
    pub fn push(&mut self, si: SectionInfo) {
        self.data_size += si.data_size;
        self.sections_info.push(si);
    }

    pub fn iter(&self) -> impl Iterator<Item = &SectionInfo> {
        self.sections_info.iter()
    }

    pub fn data_size(&self) -> u64 {
        self.data_size
    }
}

// TODO: use attributes for field section names, indexes etc.
#[derive(Default)]
pub struct Loadable {
    text: SegmentSections,
    rodata: SegmentSections,
    data: SegmentSections,
    bss: SegmentSections,
}

impl Loadable {
    pub fn process_section(&mut self, section: Section) -> Result<(), Error> {
        match section {
            Section::Text(si) => self.text.push(si),
            Section::Data(si, true) => self.rodata.push(si),
            Section::Data(si, false) => self.data.push(si),
            Section::Bss(si) => self.bss.push(si),
        }

        Ok(())
    }

    pub const fn section_names() -> &'static [&'static str] {
        &[".text", ".rodata", ".data", ".bss"]
    }

    pub const fn section_count() -> usize {
        Self::section_names().len()
    }

    pub fn segment_sections(&self) -> impl Iterator<Item = &SegmentSections> {
        let text = once(&self.text);
        let rodata = once(&self.rodata);
        let data = once(&self.data);
        let bss = once(&self.bss);

        text.chain(rodata).chain(data).chain(bss)
    }

    pub fn program_bytes<'l, R: Relocatable>(
        &'l self,
        objects: &'l [R],
    ) -> impl Iterator<Item = Result<&'l [u8], Error>> {
        let text_iter = self.text.iter();
        let rodata_iter = self.rodata.iter();

        let data1 = text_iter.filter_map(move |si| match si.data_index {
            None => None,
            Some(di) => Some(objects[di.object_index].section_data(di)),
        });
        let data2 = rodata_iter.filter_map(move |si| match si.data_index {
            None => None,
            Some(di) => Some(objects[di.object_index].section_data(di)),
        });

        data1.chain(data2)
    }

    pub fn data_bytes<'l, R: Relocatable>(
        &'l self,
        objects: &'l [R],
    ) -> impl Iterator<Item = Result<&'l [u8], Error>> {
        let iter = self.data.iter().filter_map(move |si| match si.data_index {
            None => None,
            Some(di) => Some(objects[di.object_index].section_data(di)),
        });

        iter
    }

    // .text + .rodata
    pub fn program_size(&self) -> u64 {
        let text_iter = self.text.iter();
        let rodata_iter = self.rodata.iter();

        let mut result = 0u64;
        for section in text_iter.chain(rodata_iter) {
            result += section.data_size
        }

        result
    }

    // data
    pub fn data_size(&self) -> u64 {
        let mut result = 0u64;
        for section in self.data.iter() {
            result += section.data_size
        }

        result
    }

    // bss
    pub fn bss_size(&self) -> u64 {
        let mut result = 0u64;
        for section in self.bss.iter() {
            result += section.data_size
        }

        result
    }
}
